import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.Random;

public class SyntheticData {
	public static void main(String[] args) {
		try {
			String file = args[0];
			String fileLabel = args[1];
			int nSegments = Integer.parseInt(args[2]);
			int sizeSegment = Integer.parseInt(args[3]);
			int xdim = Integer.parseInt(args[4]);
			int ydim = Integer.parseInt(args[5]);
			double sigmanoise = Double.parseDouble(args[6]);
			
			//createDataWithComplexChanges(file, nSegments, sizeSegment, xdim, ydim, sigmanoise);
			createDataWithLinearChanges(file, fileLabel, nSegments, sizeSegment, xdim, ydim, sigmanoise);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	// assumption: ydim = xdim
	private static void createDataWithComplexChanges(String file, String fileLabel, int nSegments, int sizeSegment, int xdim, int ydim, double sigmanoise) throws Exception {
		BufferedWriter writer = new BufferedWriter(new FileWriter(new File(file)));
		BufferedWriter writerLabel = new BufferedWriter(new FileWriter(new File(fileLabel)));
		
		for (int i = 0; i < nSegments; i++) {
			createSinglSegment(writer, writerLabel, i, sizeSegment, xdim, ydim, i % 4, sigmanoise);
			System.out.printf("segment %d\n", i);
		}
		
		writer.flush();
		writer.close();
		writerLabel.flush();
		writerLabel.close();
	}
	
	// assumption: ydim = xdim
	private static void createDataWithLinearChanges(String file, String fileLabel, int nSegments, int sizeSegment, int xdim, int ydim, double sigmanoise) throws Exception {
		BufferedWriter writer = new BufferedWriter(new FileWriter(new File(file)));
		BufferedWriter writerLabel = new BufferedWriter(new FileWriter(new File(fileLabel)));
		
		for (int i = 0; i < nSegments; i++) {
			createSinglSegmentLinear(writer, writerLabel, i, sizeSegment, xdim, ydim, i % 2, sigmanoise);
			System.out.printf("segment %d\n", i);
		}
		
		writer.flush();
		writer.close();
		writerLabel.flush();
		writerLabel.close();
	}
	
	private static void createSinglSegment(BufferedWriter writer, BufferedWriter writerLabel, int segmentId, int sizeSegment, int xdim, int ydim, int type, double sigmanoise) throws Exception {
		Random r = new Random();
		
		double[] data = new double[xdim + ydim];
		
		int xdimUsed = 200;
		double[] z = new double[xdimUsed];
		double[] u = new double[xdimUsed];
		for (int i = 0; i < sizeSegment; i++) {
			for (int j = 0; j < xdimUsed; j++) {
				z[j] = r.nextGaussian();
			}
			
			for (int j = 0; j < xdimUsed; j++) {
				data[j] = 0;
				for (int k = 0; k < xdimUsed; k++) {
					data[j] += r.nextDouble() * z[k];
				}
				u[j] = data[j];
			}
			
			for (int j = xdimUsed; j < xdim; j++) {
				data[j] = r.nextGaussian();
			}
			
			/*for (int j = 0; j < xdimUsed; j++) {
				u[j] = 0;
				for (int k = 0; k < xdimUsed; k++) {
					u[j] += r.nextDouble() * 0.5 * data[k];
				}
			}*/
						
			for (int j = 0; j < xdimUsed; j++) {
				if (type == 0) {
					data[j + xdim] = u[j] * u[j] - 2 * u[j] + r.nextGaussian() * sigmanoise;
				} else if (type == 1) {
					data[j + xdim] = u[j] * u[j] * u[j] + 3 * u[j] + 1 + r.nextGaussian() * sigmanoise;
				} else if (type == 2) {
					data[j + xdim] = Math.log(Math.abs(u[j] + 1)) + r.nextGaussian() * sigmanoise;
				} else if (type == 3) {
					data[j + xdim] = Math.sin(2 * u[j]) + r.nextGaussian() * sigmanoise;
				}
				/*else if (type == 4) {
					data[j + xdim] = Math.sin(u[j]) + 0.333 * (Math.tanh(2 * u[j]) + Math.tanh(3 * u[j] + 1)
							+ Math.tanh(4 * u[j] + 2)) + r.nextGaussian() * sigmanoise;
				} else {
					data[j + xdim] = Math.pow(u[j], 2) + r.nextGaussian() * sigmanoise;
				}*/
			}
			
			for (int j = xdim + xdimUsed; j < data.length; j++) {
				data[j] = r.nextGaussian();
			}
			
			for (int j = 0; j < data.length; j++) {
				writer.write(Double.toString(data[j]) + ",");
			}
			writer.write(Integer.toString(segmentId));
			writer.newLine();
			writerLabel.write(Integer.toString(segmentId));
			writerLabel.newLine();
		}
	}
	
	private static void createSinglSegmentLinear(BufferedWriter writer, BufferedWriter writerLabel, int segmentId, int sizeSegment, int xdim, int ydim, int type, double sigmanoise) throws Exception {
		Random r = new Random();
		
		double[] data = new double[xdim + ydim];
		
		int xdimUsed = 200;
		double[] z = new double[xdimUsed];
		double[] u = new double[xdimUsed];
		for (int i = 0; i < sizeSegment; i++) {
			for (int j = 0; j < xdimUsed; j++) {
				z[j] = r.nextGaussian();
			}
			
			for (int j = 0; j < xdimUsed; j++) {
				data[j] = 0;
				for (int k = 0; k < xdimUsed; k++) {
					data[j] += r.nextDouble() * z[k];
				}
				u[j] = data[j];
			}
			
			for (int j = xdimUsed; j < xdim; j++) {
				data[j] = r.nextGaussian();
			}
			
			/*for (int j = 0; j < xdimUsed; j++) {
				u[j] = 0;
				for (int k = 0; k < xdimUsed; k++) {
					u[j] += r.nextDouble() * 0.5 * data[k];
				}
			}*/
						
			for (int j = 0; j < xdimUsed; j++) {
				if (type == 0) {
					data[j + xdim] = 2 * u[j] + 1 + r.nextGaussian() * sigmanoise;
				} else if (type == 1) {
					data[j + xdim] = u[j] / 3 - 4 + r.nextGaussian() * sigmanoise;
				}
				/*else if (type == 4) {
					data[j + xdim] = Math.sin(u[j]) + 0.333 * (Math.tanh(2 * u[j]) + Math.tanh(3 * u[j] + 1)
							+ Math.tanh(4 * u[j] + 2)) + r.nextGaussian() * sigmanoise;
				} else {
					data[j + xdim] = Math.pow(u[j], 2) + r.nextGaussian() * sigmanoise;
				}*/
			}
			
			for (int j = xdim + xdimUsed; j < data.length; j++) {
				data[j] = r.nextGaussian();
			}
			
			for (int j = 0; j < data.length; j++) {
				writer.write(Double.toString(data[j]) + ",");
			}
			writer.write(Integer.toString(segmentId));
			writer.newLine();
			writerLabel.write(Integer.toString(segmentId));
			writerLabel.newLine();
		}
	}
}