package sca.detection;

import java.util.ArrayList;

import org.apache.commons.math3.linear.ArrayRealVector;
import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.linear.RealVector;

import scd.common.CommonUtils;
import scd.constants.Constants;
import scd.correlation.Correlation;
import scd.correlation.Pair;
import scd.divergence.Quadratic;
import scd.transform.Transform;

public class ChangeDetection {
	public static double[] detection(ArrayList<double[]> data, ArrayList<Integer> labels, int m, int n, int c) throws Exception {
		ArrayList<ChangeInstant> instants = detection(data, m, n, c);
		ArrayList<Integer> results = new ArrayList<Integer>();
		int T = data.size();
		for (int i = 0; i < T; i++) {
			if (i < 2 * m - 1) {
				//results.add(labels.get(i));
				results.add(0);
			} else {
				results.add(0);
			}
		}
		
		for (ChangeInstant ins : instants) {
			results.set(ins.id, 1);
		}
		
		int np = 0;
		double correct = 0;
		double late = 0;
		double miss = 0;
		boolean[] hasChecked = new boolean[T];
		for (int i = 0; i < T; i++) {
			if (labels.get(i) == 1) {
				np++;
				boolean found = false;
				for (int j = i; j < i + Constants.DELAY * m && j < T; j++) {
					if (results.get(j) == 1) {
						hasChecked[j] = true;
						correct++;
						found = true;
						break;
					}
				}
				
				if (!found) {
					found = false;
					int upper = (int)(i + Constants.DELAY * m);
					for (int j = upper; j < i + (Constants.DELAY + 1) * m && j < T; j++) {
						if (results.get(j) == 1) {
							hasChecked[j] = true;
							late++;
							found = true;
							break;
						}
					}
					
					if (!found) {
						miss++;
					}
				}
			}
		}
		
		double wrong = 0;
		for (int i = 0; i < T; i++) {
			if (results.get(i) == 1 && !hasChecked[i]) {
				wrong++;
			}
		}
		
		double[] ret = new double[5];
		ret[0] = (correct + late) / np; // TP
		ret[1] = wrong / (T - np); // FP
		ret[2] = miss / np; // FN
		ret[3] = late / np; // late rate
		ret[4] = 2 * (correct + late) / (2 * (correct + late) + wrong + miss); // F1 = 2 * TP / (2 * TP + FP + FN)
		return ret;
	}
	
	private static ArrayList<ChangeInstant> detection(ArrayList<double[]> data, int m, int n, int c) throws Exception {
		int T = data.size();
		ArrayList<ChangeInstant> ret = new ArrayList<ChangeInstant>();
		
		ArrayList<double[]> ref = new ArrayList<double[]>();
		for (int i = 0; i < m; i++) {
			ref.add(data.get(i));
		}
		
		// center the data
		double[] mean = CommonUtils.meanVector(ref);
		CommonUtils.adjustData(ref, mean);
		
		// get global weight
		double globalWeight = getGlobalWeight(mean);
		
		// perform pca
		RealMatrix[] pcaResults = null;
		if (Constants.METHOD == Constants.LIGHT || Constants.METHOD == Constants.LIGHTFULL || Constants.METHOD == Constants.LIGHTIND) {
			pcaResults = Transform.fastPCA(ref, c);
		} else if (Constants.METHOD == Constants.PCAFULL || Constants.METHOD == Constants.PCAIND) {
			pcaResults = Transform.pca(ref);
		}
		
		// transform reference window
		int k = n;
		RealMatrix transformRef = null;
		RealMatrix V = null;
		if (pcaResults != null) {
			transformRef = pcaResults[0];
			k = transformRef.getColumnDimension();
			V = pcaResults[1];
			ref.clear();
			for (int i = 0; i < m; i++) {
				ref.add(transformRef.getRow(i));
			}
		}
		
		int[] dim = new int[k];
		for (int i = 0; i < k; i++) {
			dim[i] = i;
		}
		
		// factorize joint distribution
		ArrayList<Pair> pairs = new ArrayList<Pair>();
		if (Constants.METHOD == Constants.LIGHT || Constants.METHOD == Constants.FULLFACT) {
			pairs = Correlation.extractPairs(ref);
		}
		
		// get degree of dimensions
		double[] dimWeight = null;
		if ((Constants.METHOD == Constants.LIGHT || Constants.METHOD == Constants.FULLFACT) && k > 1) {
			dimWeight = Quadratic.computeDimWeight(pairs, k);
		}
		
		// pre-compute terms 1 of quadratic score
		ArrayList<double[]> tmp = null; 
		double[] term1pair = null;
		double[] term1single = null;
		double term1 = 0;
		if (Constants.METHOD == Constants.LIGHT || Constants.METHOD == Constants.LIGHTIND || Constants.METHOD == Constants.PCAIND || Constants.METHOD == Constants.FULLFACT) {
			tmp = Quadratic.computeTerm1(ref, pairs, dimWeight, k, globalWeight / 2);
			term1pair = tmp.get(0);
			term1single = tmp.get(1);
		} else {
			term1 = Quadratic.computeTerm1(ref, globalWeight / 2);
		}
		
		ArrayList<Double> contr = null;
		ArrayList<ArrayList<Double>> inter = null;
		double changeScore = Double.NEGATIVE_INFINITY;
		//double changeScore2 = 0;
		RealVector newVector = null;
		double[] newPoint = null;
		boolean first = true;
		RealMatrix orgTest;
		RealMatrix transformTest;
		ArrayList<double[]> test = null;
		double[] val = new double[3];
		double[] adjustedVector = null;
		int t = 0;
		for (int i = 2 * m - 1; i < T; i++) {
			if (first) { // first test window
				if (transformRef != null) {
					// form the test window
					orgTest = transformRef.createMatrix(m, n);
					for (int j = m - 1; j >= 0; j--) {
						adjustedVector = CommonUtils.adjustVector(data.get(i - j), mean);
						orgTest.setRow(m - 1 - j, adjustedVector);
					}
					
					// transform it
					transformTest = orgTest.multiply(V);
					test = new ArrayList<double[]>();
					for (int j = 0; j < m; j++) {
						test.add(transformTest.getRow(j));
					}
				} else {
					test = new ArrayList<double[]>();
					for (int j = m - 1; j >= 0; j--) {
						adjustedVector = CommonUtils.adjustVector(data.get(i - j), mean);
						test.add(adjustedVector);
					}
				}
				
				// compute change score
				contr = new ArrayList<Double>();
				inter = new ArrayList<ArrayList<Double>>();
				if (Constants.METHOD == Constants.LIGHT || Constants.METHOD == Constants.FULLFACT) {
					changeScore = Quadratic.computeDiv(test, ref, pairs, dimWeight, term1pair, term1single, contr, inter, globalWeight, k, globalWeight / 2);
				} else if (Constants.METHOD == Constants.FULL || Constants.METHOD == Constants.LIGHTFULL || Constants.METHOD == Constants.PCAFULL) {
					changeScore = Quadratic.computeScoreFull(test, ref, dim, contr, inter, term1, true, globalWeight / 2);
					//changeScore2 = Quadratic.computeDiv(ref, test, globalWeight / 2);
				} else if (Constants.METHOD == Constants.LIGHTIND || Constants.METHOD == Constants.PCAIND) {
					changeScore = Quadratic.computeDivInd(test, ref, term1single, contr, inter, k, globalWeight / 2);
				} else {
					changeScore = -1;
				}
				
				// reset statistics to assess score significance
				val[0] = 0;
				val[1] = 0;
				val[2] = 0;
				t = 1;
				first = false;
			} else { // subsequent test point
				// center new test point
				adjustedVector = CommonUtils.adjustVector(data.get(i), mean);
				
				// map new test point to new space
				if (transformRef != null) {
					newVector = new ArrayRealVector(adjustedVector);
					newPoint = V.preMultiply(newVector).toArray();
				} else {
					newPoint = adjustedVector;
				}
				
				// compute new change score
				if (Constants.METHOD == Constants.LIGHT || Constants.METHOD == Constants.FULLFACT) {
					changeScore = Quadratic.updateDiv(test, ref, newPoint, pairs, dimWeight, term1pair, term1single, contr, inter, changeScore, globalWeight, k, globalWeight / 2);
				} else if (Constants.METHOD == Constants.FULL || Constants.METHOD == Constants.LIGHTFULL || Constants.METHOD == Constants.PCAFULL) {
					changeScore = Quadratic.updateScoreFull(test, ref, newPoint, dim, contr, inter, changeScore, term1, true, globalWeight / 2);
				} else if (Constants.METHOD == Constants.LIGHTIND || Constants.METHOD == Constants.PCAIND) {
					changeScore = Quadratic.updateDivInd(test, ref, newPoint, term1single, contr, inter, changeScore, k, globalWeight / 2);
				} else {
					changeScore = -1;
				}
				t++;
			}
			
			if (changeScore < 0) {
				if (Math.abs(changeScore) < Constants.EPSILON) {
					changeScore = 0;
				} else {
					throw new IllegalArgumentException(String.format("illegal score: %f\n", changeScore));
				}
			}
			if (isSignificant(changeScore, val[0], val[1], val[2], t, val)) { // assess if score is significant
				ret.add(new ChangeInstant(i, changeScore));
				
				if (T - i <=  2 * m) {
					break;
				}
				
				// form new reference window
				ref.clear();
				for (int j = i + 1; j < m + i + 1; j++) {
					ref.add(data.get(j));
				}
				
				// center it
				mean = CommonUtils.meanVector(ref);
				CommonUtils.adjustData(ref, mean);
				
				// re-compute global weight
				globalWeight = getGlobalWeight(mean);
				
				// pca transform the window
				pcaResults = null;
				if (Constants.METHOD == Constants.LIGHT || Constants.METHOD == Constants.LIGHTFULL || Constants.METHOD == Constants.LIGHTIND) {
					pcaResults = Transform.fastPCA(ref, c);
				} else if (Constants.METHOD == Constants.PCAFULL || Constants.METHOD == Constants.PCAIND) {
					pcaResults = Transform.pca(ref);
				}
				
				// transform the reference window
				k = n;
				transformRef = null;
				V = null;
				if (pcaResults != null) {
					transformRef = pcaResults[0];
					k = transformRef.getColumnDimension();
					V = pcaResults[1];
					ref.clear();
					for (int j = 0; j < m; j++) {
						ref.add(transformRef.getRow(j));
					}
				}
				
				dim = new int[k];
				for (int j = 0; j < k; j++) {
					dim[j] = j;
				}
				
				// factorize joint distribution
				pairs = new ArrayList<Pair>();
				if (Constants.METHOD == Constants.LIGHT || Constants.METHOD == Constants.FULLFACT) {
					pairs = Correlation.extractPairs(ref);
				}
				
				// compute degree of dimensions
				dimWeight = null;
				if ((Constants.METHOD == Constants.LIGHT || Constants.METHOD == Constants.FULLFACT) && k > 1) {
					dimWeight = Quadratic.computeDimWeight(pairs, k);
				}
				
				// pre-compute terms 1 for quadratic scores
				tmp = null; 
				term1pair = null;
				term1single = null;
				term1 = 0;
				if (Constants.METHOD == Constants.LIGHT || Constants.METHOD == Constants.LIGHTIND || Constants.METHOD == Constants.PCAIND || Constants.METHOD == Constants.FULLFACT) {
					tmp = Quadratic.computeTerm1(ref, pairs, dimWeight, k, globalWeight / 2);
					term1pair = tmp.get(0);
					term1single = tmp.get(1);
				} else {
					term1 = Quadratic.computeTerm1(ref, globalWeight / 2);
				}
				
				// jump to the next test point
				i = 2 * m + i - 1;
				first = true;
			}
		}
		
		return ret;
	}
	
	public static double getGlobalWeight(double[] mean) {
		double MAXVAL = Constants.MAXVAL;
		int cols = mean.length;
		double sum = 0;
		for (int i = 0; i < cols; i++) {
			sum += Math.max(mean[i] * mean[i], (MAXVAL - mean[i]) * (MAXVAL - mean[i]));
		}
		
		return 2 * Math.sqrt(sum);
	}
	
	public static boolean isSignificant(double score, double pastMean, double pastSum, double M_current, int t, double[] vals) {
		double p_mean_current = (pastMean * (t - 1) + score) / t;
		double m_current = pastSum + p_mean_current - score + Constants.DELTA;
		if (Math.abs(m_current) > M_current) {
			M_current = Math.abs(m_current);
		}
		vals[0] = p_mean_current;
		vals[1] = m_current;
		vals[2] = M_current;
		
		if (M_current - m_current > Constants.FACTOR * p_mean_current) {
			return true;
		}
		
		return false;
	}
}