package scd.common;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Scanner;

import scd.constants.Constants;

public class CommonUtils {
	public static ArrayList<double[]> readData(Scanner scan, double[] mean, int m, int n) {
		ArrayList<double[]> ref = new ArrayList<double[]>();
		String[] vals = null;
		double[] tmp = null;
		for (int i = 0; i < m; i++) {
			vals = scan.nextLine().split(",");
			tmp = new double[n];
			for (int j = 0; j < n; j++) {
				tmp[j] = Double.parseDouble(vals[j]);
			}
			
			if (mean == null) {
				ref.add(tmp);
			} else {
				ref.add(adjustVector(tmp, mean));
			}
		}
		
		return ref;
	}
	
	public static ArrayList<double[]> readData(String file, int cols, ArrayList<Integer> labels) throws IOException {
		ArrayList<double[]> data = new ArrayList<double[]>();
		
		double[] min = new double[cols];
		double[] max = new double[cols];
		for (int i = 0; i < cols; i++) {
			min[i] = Double.POSITIVE_INFINITY;
			max[i] = Double.NEGATIVE_INFINITY;
		}
		
		Scanner scan = new Scanner(new File(file));
		String[] vals = null;
		double[] a = null;
		while (scan.hasNext()) {
			vals = scan.nextLine().split(",");
			if (vals.length != cols + 1) {
				scan.close();
				throw new IllegalArgumentException(String.format("number of columns mismatch: cols = %d, vals.length = %d", cols, vals.length));
			}
			
			a = new double[cols];
			for (int i = 0; i < cols; i++) {
				a[i] = Double.parseDouble(vals[i]);
				
				if (a[i] > max[i]) {
					max[i] = a[i];
				}
				
				if (a[i] < min[i]) {
					min[i] = a[i];
				}
			}
			
			data.add(a);
			labels.add(Integer.parseInt(vals[cols]));
		}
		scan.close();
		
		// normalize data
		int rows = data.size();
		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++) {
				if (min[j] == max[j]) {
					data.get(i)[j] = 0;
				} else {
					data.get(i)[j] = Constants.MAXVAL * (data.get(i)[j] - min[j]) / (max[j] - min[j]);
				}
			}
		}
		
		return data;
	}
	
	public static int readDataLabels(String file, int cols, ArrayList<Integer> labels) throws IOException {
		Scanner scan = new Scanner(new File(file));
		String s = null;
		String[] vals = null;
		int T = 0;
		while (scan.hasNext()) {
			s = scan.nextLine();
			if (s.isEmpty()) {
				break;
			}
			
			vals = s.split(",");
			if (vals.length != cols + 1) {
				scan.close();
				throw new IllegalArgumentException(String.format("number of columns mismatch: cols = %d, vals.length = %d", cols, vals.length));
			}
			
			labels.add(Integer.parseInt(vals[cols]));
			T++;
		}
		scan.close();
		
		return T;
	}
	
	public static int readDataLabels(String file, ArrayList<Integer> labels) throws IOException {
		Scanner scan = new Scanner(new File(file));
		String s = null;
		int T = 0;
		while (scan.hasNext()) {
			s = scan.nextLine();
			if (s.isEmpty()) {
				break;
			}
			
			labels.add(Integer.parseInt(s));
			T++;
		}
		scan.close();
		
		return T;
	}
	
	public static void normalizeData(String fileIn, String fileOut, int cols) throws IOException {
		double[] min = new double[cols];
		double[] max = new double[cols];
		for (int i = 0; i < cols; i++) {
			min[i] = Double.POSITIVE_INFINITY;
			max[i] = Double.NEGATIVE_INFINITY;
		}
		
		Scanner scan = new Scanner(new File(fileIn));
		String s = null;
		String[] vals = null;
		double[] a = null;
		while (scan.hasNext()) {
			s = scan.nextLine();
			if (s.isEmpty()) {
				break;
			}
			
			vals = s.split(",");
			if (vals.length != cols + 1) {
				scan.close();
				throw new IllegalArgumentException(String.format("number of columns mismatch: cols = %d, vals.length = %d", cols, vals.length));
			}
			
			a = new double[cols];
			for (int i = 0; i < cols; i++) {
				a[i] = Double.parseDouble(vals[i]);
				
				if (a[i] > max[i]) {
					max[i] = a[i];
				}
				
				if (a[i] < min[i]) {
					min[i] = a[i];
				}
			}
		}
		scan.close();
		
		// normalize data
		scan = new Scanner(new File(fileIn));
		BufferedWriter writer = new BufferedWriter(new FileWriter(new File(fileOut)));
		while (scan.hasNext()) {
			s = scan.nextLine();
			if (s.isEmpty()) {
				break;
			}
			
			vals = s.split(",");
			if (vals.length != cols + 1) {
				scan.close();
				writer.flush();
				writer.close();
				throw new IllegalArgumentException(String.format("number of columns mismatch: cols = %d, vals.length = %d", cols, vals.length));
			}
			
			for (int i = 0; i < cols; i++) {
				double tmp = Double.parseDouble(vals[i]);
				
				if (min[i] == max[i]) {
					tmp = 0;
				} else {
					tmp = Constants.MAXVAL * (tmp - min[i]) / (max[i] - min[i]);
				}
				
				writer.write(Double.toString(tmp) + ",");
			}
			writer.write(vals[cols]);
			writer.newLine();
		}
		scan.close();
		writer.flush();
		writer.close();
	}
	
	public static double[] meanVector(ArrayList<double[]> data) {
		int rows = data.size();
		int cols = data.get(0).length;
		double[] mean = new double[cols];
		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++) {
				mean[j] += data.get(i)[j] / rows;
			}
		}
		
		return mean;
	}
	
	public static void adjustData(ArrayList<double[]> data, double[] mean) {
		int rows = data.size();
		int cols = data.get(0).length;
		for (int i = 0; i < rows; i++) {
			for (int j = 0; j < cols; j++) {
				data.get(i)[j] = data.get(i)[j] - mean[j];
			}
		}
	}
	
	public static double[] adjustVector(double[] data, double[] mean) {
		int cols = data.length;
		for (int j = 0; j < cols; j++) {
			data[j] = data[j] - mean[j];
		}
		
		return data;
	}
	
	public static void readInputString(String[] args) throws Exception
	{	
		int i;
		int total = args.length - 1;
		
		// take FILEIN
		boolean found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-FILEIN"))
			{
				Constants.FILEIN = args[i + 1];
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -FILEIN");
		
		// take FILELABEL
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-FILELABEL"))
			{
				Constants.FILELABEL = args[i + 1];
				found = true;
				break;
			}
		
		// take FILEOUT
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-FILEOUT"))
			{
				Constants.FILEOUT = args[i + 1];
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -FILEOUT");
		
		// take DIM
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-DIM"))
			{
				Constants.DIM = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -DIM");
		
		// take WSIZE
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-WSIZE"))
			{
				Constants.WSIZE = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
		if (found == false)
			throw new Exception("Missing -WSIZE");
		
		// take SUBDIM
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-SUBDIM"))
			{
				Constants.SUBDIM = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
		if (found == false) {
			Constants.SUBDIM = Constants.DIM;
		}
		
		// take S1
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-S1"))
			{
				Constants.S1 = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
		
		// take S2
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-S2"))
			{
				Constants.S2 = Integer.parseInt(args[i + 1]);
				found = true;
				break;
			}
		
		// take PERCENT
		found = false;
		for (i = 0; i < total; i++)
			if (args[i].equals("-PERCENT"))
			{
				Constants.PERCENT = Double.parseDouble(args[i + 1]);
				found = true;
				break;
			}
	} // end method
}