package scd.divergence;

import java.util.ArrayList;

import scd.correlation.Pair;

public class Quadratic {
	public static double[] computeDimWeight(ArrayList<Pair> pairs, int k) {
		double[] weight = new double[k];
		for (Pair p : pairs) {
			weight[p.x] = weight[p.x] + 1;
			weight[p.y] = weight[p.y] + 1;
		}

		return weight;
	}

	public static ArrayList<double[]> computeTerm1(ArrayList<double[]> ref,
			ArrayList<Pair> pairs, double[] dimWeight, int k, double max) {
		double[] term1pair = new double[k - 1];
		double[] term1single = new double[k];

		int idx = 0;
		for (Pair p : pairs) {
			term1pair[idx] = computeTerm1Pair(ref, p, max);
			idx++;
		}

		for (int i = 0; i < k; i++) {
			if (dimWeight == null || dimWeight[i] > 1) {
				term1single[i] = computeTerm1Single(ref, i, max);
			}
		}

		ArrayList<double[]> ret = new ArrayList<double[]>();
		ret.add(term1pair);
		ret.add(term1single);

		return ret;
	}
	
	public static double computeTerm1(ArrayList<double[]> ref, double max) {
		int N = ref.size();
		int d = ref.get(0).length;
		
		double term1 = 0;
		for (int i = 0; i < N; i++) {
			for (int j = i; j < N; j++) {
				double prod = 1;
				for (int k = 0; k < d; k++) {
					prod *= (max - Math.max(ref.get(i)[k], ref.get(j)[k]));
				}
				
				if (j != i) {
					term1 += 2 * prod / (N * N);
				} else {
					term1 += prod / (N * N);
				}
			}
		}

		return term1;
	}
	
	public static double computeDiv(ArrayList<double[]> ref, ArrayList<double[]> test, double max) {
		int N = ref.size();
		int T = test.size();
		int d = ref.get(0).length;
		
		double term1 = 0;
		for (int i = 0; i < N; i++) {
			for (int j = i; j < N; j++) {
				double prod = 1;
				for (int k = 0; k < d; k++) {
					prod *= (max - Math.max(ref.get(i)[k], ref.get(j)[k]));
				}
				
				if (j != i) {
					term1 += 2 * prod / (N * N);
				} else {
					term1 += prod / (N * N);
				}
			}
		}
		
		double term2 = 0;
		for (int i = 0; i < N; i++) {
			for (int j = 0; j < T; j++) {
				double prod = 1;
				for (int k = 0; k < d; k++) {
					prod *= (max - Math.max(ref.get(i)[k], test.get(j)[k]));
				}
				
				term2 -= 2 * prod / (N * T);
			}
		}
		
		double term3 = 0;
		for (int i = 0; i < T; i++) {
			for (int j = i; j < T; j++) {
				double prod = 1;
				for (int k = 0; k < d; k++) {
					prod *= (max - Math.max(test.get(i)[k], test.get(j)[k]));
				}
				
				if (j != i) {
					term3 += 2 * prod / (T * T);
				} else {
					term3 += prod / (T * T);
				}
			}
		}

		return term1 + term2 + term3;
	}
	
	private static double computeTerm1Pair(ArrayList<double[]> ref, Pair p, double max) {
		int N = ref.size();
		
		double term1 = 0;
		for (int i = 0; i < N; i++) {
			for (int j = i; j < N; j++) {
				double prod = 1;
				prod *= (max - Math.max(ref.get(i)[p.x], ref.get(j)[p.x]));
				prod *= (max - Math.max(ref.get(i)[p.y], ref.get(j)[p.y]));
				
				if (j != i) {
					term1 += 2 * prod / (N * N);
				} else {
					term1 += prod / (N * N);
				}
			}
		}

		return term1;
	}

	private static double computeTerm1Single(ArrayList<double[]> ref, int dim, double max) {
		int N = ref.size();
		
		double term1 = 0;
		for (int i = 0; i < N; i++) {
			for (int j = i; j < N; j++) {
				double prod = (max - Math.max(ref.get(i)[dim], ref.get(j)[dim]));
				
				if (j != i) {
					term1 += 2 * prod / (N * N);
				} else {
					term1 += prod / (N * N);
				}
			}
		}

		return term1;
	}

	public static double computeDiv(ArrayList<double[]> test,
			ArrayList<double[]> ref, ArrayList<Pair> pairs, double[] dimWeight,
			double[] term1pair, double[] term1single,
			ArrayList<Double> contr, ArrayList<ArrayList<Double>> inter,
			double globalWeight, int k, double max) {
		contr.clear();
		inter.clear();
		int N = test.size();
		for (int i = 0; i < N; i++) {
			inter.add(new ArrayList<Double>());
		}
		System.out.printf("inter.size() = %d\n", inter.size());
		
		double score = 0;
		
		if (k > 1) {
			int idx = 0;
			for (Pair p : pairs) {
				int[] dim = new int[2];
				dim[0] = p.x;
				dim[1] = p.y;
	
				boolean init = false;
				if (idx == 0) {
					init = true;
				}
	
				double a = computeScore(test, ref, dim, contr, inter, term1pair[idx], globalWeight, init, max);
				score += a;
				idx++;
			}
	
			for (int i = 0; i < k; i++) {
				if (dimWeight[i] > 1) {
					int[] dim = new int[1];
					dim[0] = i;
					double a = computeScore(test, ref, dim, contr, inter, term1single[i], 1 - dimWeight[i], false, max);
					score -= a;
				}
			}
		} else {
			int[] dim = new int[1];
			dim[0] = 0;
			score = computeScore(test, ref, dim, contr, inter, term1single[0], 1, true, max);
		}
		
		//System.out.printf("k = %d\n", k);
		//System.out.printf("pairs.size() = %d\n", pairs.size());
		//System.out.printf("contr.size() = %d\n\n", contr.size());

		return score;
	}
	
	public static double computeScoreFull(ArrayList<double[]> test,
			ArrayList<double[]> ref, int[] dim, ArrayList<Double> contr,
			ArrayList<ArrayList<Double>> inter, double term1,
			boolean init, double max) {
		contr.clear();
		inter.clear();
		int N = test.size();
		for (int i = 0; i < N; i++) {
			inter.add(new ArrayList<Double>());
		}
		System.out.printf("inter.size() = %d\n", inter.size());

		double score = computeScore(test, ref, dim, contr, inter, term1, 1, true, max);
		return score;
	}
	
	public static double computeDivInd(ArrayList<double[]> test, ArrayList<double[]> ref,
			double[] term1single, ArrayList<Double> contr, ArrayList<ArrayList<Double>> inter, int k, double max) {
		contr.clear();
		inter.clear();
		int N = test.size();
		for (int i = 0; i < N; i++) {
			inter.add(new ArrayList<Double>());
		}
		System.out.printf("inter.size() = %d\n", inter.size());

		double score = 0;
		for (int i = 0; i < k; i++) {
			int[] dim = new int[1];
			dim[0] = i;
			
			boolean init = false;
			if (i == 0) {
				init = true;
			}
			
			double a = computeScore(test, ref, dim, contr, inter, term1single[i], 1, init, max);
			score += a;
		}

		return score;
	}

	public static double updateDiv(ArrayList<double[]> test, ArrayList<double[]> ref, double[] newPoint,
			ArrayList<Pair> pairs, double[] dimWeight, double[] term1pair, double[] term1single,
			ArrayList<Double> contr, ArrayList<ArrayList<Double>> inter,
			double oldScore, double globalWeight, int k, double max) {
		int T = test.size();
		
		//System.out.printf("contr.size() = %d\n", contr.size());
		if (contr.size() != T) {
			throw new IllegalArgumentException(String.format("something wrong: contr.size() = %d", contr.size()));
		}
		
		// update score
		oldScore -= contr.get(0);
		contr.remove(0);
		for (int i = 0; i < T; i++) {
			oldScore -= inter.get(0).get(i);
			if (i != 0) {
				oldScore -= inter.get(i).get(0);
				inter.get(i).remove(0);
			}
		}
		inter.remove(0);
		inter.add(new ArrayList<Double>());
		
		test.remove(0);
		test.add(newPoint);
		
		if (k > 1) {
			int idx = 0;
			for (Pair p : pairs) {
				int[] dim = new int[2];
				dim[0] = p.x;
				dim[1] = p.y;
	
				boolean init = false;
				if (idx == 0) {
					init = true;
				}
	
				oldScore = updateScore(test, ref, newPoint, dim, contr, inter, oldScore, globalWeight, init, max);
				idx++;
			}
	
			for (int i = 0; i < k; i++) {
				if (dimWeight[i] > 1) {
					int[] dim = new int[1];
					dim[0] = i;
					oldScore = updateScore(test, ref, newPoint, dim, contr, inter, oldScore, 1 - dimWeight[i], false, max);
				}
			}
		} else {
			int[] dim = new int[1];
			dim[0] = 0;
			oldScore = updateScore(test, ref, newPoint, dim, contr, inter, oldScore, 1, true, max);
		}
		
		return oldScore;
	}
	
	public static double updateScoreFull(ArrayList<double[]> test,
			ArrayList<double[]> ref, double[] newPoint, int[] dim,
			ArrayList<Double> contr, ArrayList<ArrayList<Double>> inter,
			double oldScore, double term1, boolean init, double max) {
		int T = test.size();
		
		// update score
		oldScore -= contr.get(0);
		contr.remove(0);
		for (int i = 0; i < T; i++) {
			oldScore -= inter.get(0).get(i);
			if (i != 0) {
				oldScore -= inter.get(i).get(0);
				inter.get(i).remove(0);
			}
		}
		inter.remove(0);
		inter.add(new ArrayList<Double>());
		
		test.remove(0);
		test.add(newPoint);
		
		oldScore = updateScore(test, ref, newPoint, dim, contr, inter, oldScore, 1, true, max);
		return oldScore;
	}
	
	public static double updateDivInd(ArrayList<double[]> test, ArrayList<double[]> ref, double[] newPoint,
			double[] term1single, ArrayList<Double> contr, ArrayList<ArrayList<Double>> inter,
			double oldScore, int k, double max) {
		int T = test.size();
		
		// update score
		oldScore -= contr.get(0);
		contr.remove(0);
		for (int i = 0; i < T; i++) {
			oldScore -= inter.get(0).get(i);
			if (i != 0) {
				oldScore -= inter.get(i).get(0);
				inter.get(i).remove(0);
			}
		}
		inter.remove(0);
		inter.add(new ArrayList<Double>());
		
		test.remove(0);
		test.add(newPoint);
		
		for (int i = 0; i < k; i++) {
			int[] dim = new int[1];
			dim[0] = i;
			
			boolean init = false;
			if (i == 0) {
				init = true;
			}
			
			oldScore = updateScore(test, ref, newPoint, dim, contr, inter, oldScore, 1, init, max);
		}
		
		return oldScore;
	}
	
	public static double computeScore(ArrayList<double[]> test,
			ArrayList<double[]> ref, int[] dim, ArrayList<Double> contr,
			ArrayList<ArrayList<Double>> inter, double term1, double weight,
			boolean init, double max) {
		int T = test.size();
		int N = ref.size();
		int d = dim.length;
		
		//System.out.printf("init = %b\n\n", init);
		double s = term1;
		for (int i = 0; i < T; i++) {
			double si = 0;
			for (int j = 0; j < N; j++) {
				double prod = 1;
				for (int k = 0; k < d; k++) {
					prod *= (max - Math.max(test.get(i)[dim[k]], ref.get(j)[dim[k]]));
				}

				si -= 2 * prod / (T * N);
			}
			if (init) {
				contr.add(weight * si);
			} else {
				contr.set(i, contr.get(i) + weight * si);
			}
			s += si;

			for (int j = i; j < T; j++) {
				double prod = 1;
				for (int k = 0; k < d; k++) {
					prod *= (max - Math.max(test.get(i)[dim[k]], test.get(j)[dim[k]]));
				}

				prod = prod / (T * T);
				if (init) {
					inter.get(i).add(weight * prod);
				} else {
					inter.get(i).set(j, inter.get(i).get(j) + weight * prod);
				}
				if (j != i) {
					if (init) {
						inter.get(j).add(weight * prod);
					} else {
						inter.get(j).set(i, inter.get(j).get(i) + weight * prod);
					}
					
					s += 2 * prod;
				} else {
					s += prod;
				}
			}
		}

		return weight * s;
	}

	// assumption: contri(0) has been removed, so has inter(0); oldScore has
	// been updated to these removal
	public static double updateScore(ArrayList<double[]> test,
			ArrayList<double[]> ref, double[] newPoint, int[] dim,
			ArrayList<Double> contr, ArrayList<ArrayList<Double>> inter,
			double oldScore, double weight, boolean init, double max) {
		int T = test.size();
		int N = ref.size();
		int d = dim.length;
		//oldScore += weight * term1;

		double si = 0;
		for (int j = 0; j < N; j++) {
			double prod = 1;
			for (int k = 0; k < d; k++) {
				prod *= (max - Math.max(newPoint[dim[k]], ref.get(j)[dim[k]]));
			}

			si -= 2 * prod / (T * N);
		}
		if (init) {
			contr.add(weight * si);
		} else {
			contr.set(T - 1, contr.get(T - 1) + weight * si);
		}
		oldScore += weight * si;

		for (int j = 0; j < T; j++) {
			double prod = 1;
			for (int k = 0; k < d; k++) {
				prod *= (max - Math.max(newPoint[dim[k]], test.get(j)[dim[k]]));
			}

			prod = prod / (T * T);
			if (init) {
				inter.get(j).add(weight * prod);
			} else {
				inter.get(j).set(T - 1, inter.get(j).get(T - 1) + weight * prod);
			}

			if (j != T - 1) {
				if (init) {
					inter.get(T - 1).add(weight * prod);
				} else {
					inter.get(T - 1).set(j, inter.get(T - 1).get(j) + weight * prod);
				}
				
				oldScore += 2 * weight * prod;
			} else {
				oldScore += weight * prod;
			}
		}

		return oldScore;
	}
}